#import ipdb
import numpy as np
import pandas as pd
from tqdm import tqdm
import random

from helpers import bootstrapSamples, permutatedSamples, testStatistic, calculateP, calculateFDR
from calculateOverlaps1 import calculateOverlaps1
from calculateOverlaps2 import calculateOverlaps2

def rots(data, groups, B=1000, K=None, paired=False, seed=None, a1=None, a2=None, log=None, progress=False, verbose=True):
  if isinstance(data, pd.DataFrame):
    data_val = data.values    
  if seed is not None:
    random.seed(seed)
  if data.shape[1] != len(groups):
    raise ValueError("Number of samples in the data does not match the groups.")
  if not data.index.any():
    data.index = np.arange(1, data.shape[0]+1)

  ssq = np.concatenate((
    np.arange(0, 0.21, 0.01), 
    np.arange(0.22, 1.01, 0.02),
    np.arange(1.2, 5.2, 0.2)
  ))
  N = np.concatenate([
    np.arange(1, 21) * 5,
    np.arange(11, 51) * 10,
    np.arange(21, 41) * 25, 
    np.arange(11, 1001) * 100
  ])
  
  #ipdb.set_trace(context=6)   ## BREAKPOINT

  if K is None:
    K = data.shape[0] // 4
    if (verbose):
      print("No top list size K given, using ", K)

  K = min(K, data.shape[0])
  N = N[N < K]

  if isinstance(groups[0], str):
    groups = pd.factorize(groups)[0]
    groups_levels = pd.factorize(groups)[1]
  else:
    groups_levels = None

  data = data.iloc[:, np.argsort(groups)]
  groups = np.sort(groups)
  
  for i in np.unique(groups):
    if np.any(np.sum(np.isnan(data.iloc[:, np.where(groups == i)[0]]), axis=1) >= len(np.where(groups == i)[0]) - 1):
      if groups_levels is None:
        target = i
      else:
        target = groups_levels[i]
      raise ValueError("The data matrix of group " + str(target) + " contains rows with less than two non-missing values, please remove these rows.")

  cl = groups + (1-np.min(groups))

  ## Check number of samples for paired test
  if paired:
    for i in np.unique(cl)[-1]:
      if len(np.where(cl == 1)[0]) != len(np.where(cl == i)[0]):
        raise ValueError("Uneven number of samples for paired test.")

  ## Calculate fold change
  if len(np.unique(cl)) == 2:
    if log:
      logfc = np.mean(data.iloc[:, np.where(cl==1)[0]], axis=1) - np.mean(data.iloc[:, np.where(cl==2)[0]], axis=1)
    else:
      logfc = np.mean(np.log2(data.iloc[:,np.where(cl==1)[0]]+1), axis=1) - np.mean(np.log2(data.iloc[:,np.where(cl==2)[0]]+1), axis=1)
  else:
    logfc = np.repeat(np.nan, data.shape[0])

  ## ---------------------------------------------------------------------------

  ## Bootstrap samples
  if (verbose): 
    print("Bootstrapping samples")    

  samples = bootstrapSamples(2*B, cl, paired)
  ## Permutated samples
  pSamples = permutatedSamples(len(samples), cl)

  ## Test statistics in the bootstrap (permutate) datasets.
  ## For each bootstrap (permutated) dataset, determine the signal log-ratio
  ## (D) and the standard error (S).
  D = np.empty((data.shape[0], samples.shape[0]))
  S = np.empty((data.shape[0], samples.shape[0]))
  pD = np.empty((data.shape[0], samples.shape[0]))
  pS = np.empty((data.shape[0], samples.shape[0]))

  # Progress bar  
  pb = tqdm(total=samples.shape[0])

  #ipdb.set_trace(context=6)   ## BREAKPOINT
  for i in range(samples.shape[0]):
    samples_R = np.split(samples[i, :], np.where(np.diff(cl))[0]+1)
    pSamples_R = np.split(pSamples[i, :], np.where(np.diff(cl))[0]+1)
    
    ## If a1 and a2 parameters are given, we don't need the bootstrap
    ## dataset
    if a1 == None or a2 == None:
      fit = testStatistic(paired, [data.iloc[:, x].to_numpy() for x in samples_R])
      D[:,i] = fit['d']
      S[:,i] = fit['s']
    
      
    pFit = testStatistic(paired, [data.iloc[:, x].to_numpy() for x in pSamples_R])
    pD[:,i] = pFit['d']
    pS[:,i] = pFit['s']
    
    if progress: 
      pb.update()
    
  if progress: 
    pb.close()

  ## Free up memory
  del samples
  del pSamples
    
  ## ---------------------------------------------------------------------------

  if  a1==None or a2==None:
    ## Optimize the parameters
    if verbose:
      print("Optimizing parameters")
    
    ## Calculate the reproducibility values for all the given a1-values and top
    ## list sizes in both bootstrap and permuted data and their standard
    ## deviations in the bootstrap case
    
    ## Reproducibility matrix (bootstrap data):
    ## the rows correspond to the a1-values given in ssq (+ 1 for signal
    ## log-ratio only: a1=1, a2=0), the columns correspond to the different top
    ## list sizes
    reprotable = pd.DataFrame(index=np.append(ssq, "slr"), columns = N)

    ## Reproducibility matrix (permuted data):
    ## the rows correspond to the a1-values given in ssq (+ 1 for signal
    ## log-ratio only: a1=1, a2=0), the columns correspond to the different top
    ## list sizes
    reprotable_P = pd.DataFrame(index=np.append(ssq, "slr"), columns = N)

    ## Standard deviation matrix for the reproducibility values (bootstrap
    ## data): the rows correspond to the a1-values given in ssq (+ 1 for signal
    ## log-ratio only: a1=1, a2=0), the columns correspond to the different
    ## top list sizes
    reprotable_sd = pd.DataFrame(index=np.append(ssq, "slr"),columns=N)

    # Progress bar
    if progress:
        pb = tqdm(total=len(ssq)) #txtProgressBar(min=0, max=length(ssq), style=3)      

    for i in range(len(ssq)):
      ## The overlaps between bootstrap samples. Rows correspond to different
      ## bootrsrap samples and columns corrospond to different top list size.
      ## Repeat for each parameter combination a1 and a2.
      overlaps = np.zeros((B,len(N)))
      overlaps_P = np.zeros((B,len(N)))      

      ## Call the custom c++-loop 1.
      cResults = calculateOverlaps1(D, S, pD, pS, len(D), N.astype(int), len(N),
                        ssq[i], int(B), overlaps, overlaps_P)
      
      ## Colmeans & rowMeans are a lot faster than apply
      #         reprotable[i, ] <- colMeans(overlaps)
      #         reprotable.P[i, ] <- colMeans(overlaps.P)
      
      
      reprotable.iloc[i] = np.mean(cResults["overlaps"], axis=0)
      reprotable_P.iloc[i] = np.mean(cResults["overlaps_P"], axis=0)
      
      ## Standard deviation for each column
      ## same as reprotable.sd[i, ] <- apply(overlaps, 2, sd)
      ## or just sd(overlaps), but a lot faster.
      #         reprotable.sd[i,] <- sqrt(rowSums((t(overlaps) - reprotable[i,])^2) /
      #                                     (nrow(overlaps) - 1))
      #sqrt(rowSums((t(cResults[["overlaps"]]) - reprotable[i,])^2) / (nrow(cResults[["overlaps"]]) - 1)) 
      #reprotable_sd.iloc[i] = np.std(cResults["overlaps"])
      reprotable_sd.iloc[i] = np.sqrt(np.sum((cResults["overlaps"].T - reprotable.iloc[i].values[0])**2, axis=1) / 
                                      (cResults["overlaps"].shape[0] - 1))[0]
      
      if progress:
        pb.update()      

    if progress: 
      pb.close()

    i = len(ssq)
    overlaps = np.zeros((B,len(N)))
    overlaps_P = np.zeros((B,len(N)))

    ## Call the custom c++-loop 2.
    cResults = calculateOverlaps2(D, pD, len(D), N.astype(int), len(N),
                    int(B), overlaps, overlaps_P)

    ## Free up memory
    #rm(D, S)
    del D
    del S

    #ipdb.set_trace(context=6)   ## BREAKPOINT
    reprotable.iloc[i] = np.mean(cResults["overlaps"], axis=0) #colMeans(cResults[["overlaps"]])
    reprotable_P.iloc[i] = np.mean(cResults["overlaps_P"], axis=0) #colMeans(cResults[["overlaps_P"]])
    ## Standard deviation for each column
    #sqrt(rowSums((t(cResults[["overlaps"]]) - reprotable[i,])^2) / (nrow(cResults[["overlaps"]]) - 1))
    #reprotable_sd.iloc[i] = np.std(cResults["overlaps"])
    reprotable_sd.iloc[i] = np.sqrt(np.sum((cResults["overlaps"].T - reprotable.iloc[i].values[0])**2, axis=1) / 
                                      (cResults["overlaps"].shape[0] - 1))[0]

    ## Free up memory
    del overlaps
    del overlaps_P
    del cResults

    ## -------------------------------------------------------------------------

    ## Calculate the Z-statistic and find the top list size and the
    ## (a1,a2)-combination giving the maximal Z-value
    ztable = (reprotable - reprotable_P) / reprotable_sd
    ## Rownames of ztable are c(ssq, "slr") and colnames are N

    ztable = ztable.infer_objects()

    ## Free up memory
    del reprotable_P
    del reprotable_sd
    
    sel = np.unravel_index(np.argmax(ztable[np.isfinite(ztable)]), ztable.shape) #np.where(ztable == max(ztable[is.finite(ztable)]), arr.ind=TRUE)
    ## Sel is a matrix containing the location(s) of the largest value (row,
    ## col). If the location of the largest value is not unique then nrow(sel)
    ## > 2 (length(sel) > 2)

    #ipdb.set_trace(context=6)   ## BREAKPOINT
    if len(sel) > 2:
      sel = sel[0:2]

    if sel[0] < len(reprotable)-1:
      a1 = float(reprotable.index[sel[0]])
      a2 = 1
    
    if sel[0] == len(reprotable)-1:
      a1 = 1
      a2 = 0
    
    k = int(reprotable.columns[sel[1]])
    R = reprotable.iloc[sel[0],sel[1]]
    Z = ztable.iloc[sel[0],sel[1]]
    
    ## Free up memory
    del reprotable
    
    #ipdb.set_trace(context=6)   ## BREAKPOINT
    ## Calculate the reproducibility-optimized test statistic based on the
    ## reproducibility-maximizing a1, a2 and k values and the corresponding FDR      
    #fit = testStatistic(paired, lapply(split(1:len(cl), cl), function(x) data[:,x]))
    fit = testStatistic(paired, np.split(data.to_numpy(), np.where(np.diff(cl) != 0)[0] + 1, axis=1))
    d = fit['d'] / (a1 + a2 * fit['s'])
    pD = pD/(a1 + a2 * pS)
    
    ## Free up memory
    del pS
    
    if verbose: 
      print("Calculating p-values")
    p = calculateP(d, pD)
    
    if verbose:
      print("Calculating FDR")
    FDR = calculateFDR(d, pD, progress)
    
    ## Free up memory
    del pD

    ROTS_output = {
      "data": data,
      "B": B,
      "d": d,
      "logfc": logfc,
      "p": p,
      "FDR": FDR,
      "a1": a1,
      "a2": a2,
      "k": k,
      "R": R,
      "Z": Z,
      "ztable": ztable,
      "cl": cl}
      
  else:
    ## Calculate statistic based on the given parameter values
    ## and the corresponding FDR
    #fit <- testStatistic(paired, lapply(split(1:length(cl), cl), function(x) data[,x]))
    fit = testStatistic(paired, np.split(data.to_numpy(), np.where(np.diff(cl) != 0)[0] + 1, axis=1))
    d = fit['d'] / (a1 + a2 * fit['s'])
    if verbose: 
      print("Calculating p-values")
    p = calculateP(d, pD/(a1 + a2 * pS))
    if verbose: 
      print("Calculating FDR")
    FDR = calculateFDR(d, pD/(a1 + a2 * pS), progress)
    
    ROTS_output = {
      "data": data,
      "B": B,
      "d": d,
      "logfc": logfc,
      "p": p,
      "FDR": FDR,
      "a1": a1,
      "a2": a2,
      "k": None,
      "R": None,
      "Z": None,      
      "cl": cl}       
      
  ## Define the class
  return ROTS_output


import numpy as np
from numba import njit, jit
import pandas as pd
from tqdm import tqdm

@jit(nopython=True)
def bootstrapSamples(B, labels, paired):
  samples = np.zeros((B, len(labels)))#matrix(nrow=B, ncol=length(labels))
  
  for i in range(B):
    for label in np.unique(labels):
      pos = np.where(labels == label)[0]
      #samples[i, pos] = np.random.choice(pos, size=len(pos), replace=True)            
      _samp = samples[i]
      _samp[pos] = np.random.choice(pos, size=len(pos), replace=True)
      samples[i] = _samp

  if paired:
    for i in range(B):
      for label in np.unique(labels)[1:]:
        pos = np.where(labels == label)[0]
        #samples[i, pos] = samples[i, np.where(labels == 1)[0]] + pos[0]-1
        _samp = samples[i]
        _samp[pos] = _samp[np.where(labels == 1)[0]] + pos[0]-1
        samples[i] = _samp
  
  return samples

@njit
def permutatedSamples(B, cl):
  samples = np.zeros((B, len(cl)))
  for i in range(B):
    samples[i, :] = np.random.permutation(len(cl))

  return samples


def testStatistic(paired, samples):
   # Two groups
   if len(samples)==2:
     X = samples[0]
     Y = samples[1]
     
     ## Calculates the test statistic for each row.
     ## X and Y are the data matrices of the two groups.
     ## Each row of these two matrices must contain at least TWO not NA values.
     ## Thus the "variance" always exists.  
     
     #ipdb.set_trace(context=6)   ## BREAKPOINT

     ## Row means
     mX = np.nanmean(X, axis=1, keepdims=True) #rowMeans(X, na.rm=TRUE)
     mY = np.nanmean(Y, axis=1, keepdims=True) #rowMeans(Y, na.rm=TRUE)
     
     mX[np.isnan(mX)] = 0
     mY[np.isnan(mY)] = 0

     ## Pooled standard deviations for each row
     sX = np.nansum((X - mX)**2, axis=1) #rowSums((X - mX)^2, na.rm=TRUE)
     sY = np.nansum((Y - mY)**2, axis=1) #rowSums((Y - mY)^2, na.rm=TRUE)

     sX[np.isnan(sX)] = 0
     sY[np.isnan(sY)] = 0
     
     if not paired:
       ## Number of not NA values in each row
       nX = np.sum(~np.isnan(X), axis=1)
       nY = np.sum(~np.isnan(Y), axis=1)             

       ## d == difference between the group means for each row (==gene)
       ## s == pooled standard deviation for each row (==gene)        
       d = mY - mX
       s = np.sqrt(((sX + sY) / (nX + nY - 2)) * (1 / nX + 1 / nY))
       
       ## Cases with less than two non-missing values.
       ## Set d = 0, s = 1
       ind = np.where( (nY < 2) | (nX < 2) )[0]
       d[ind] = 0
       s[ind] = 1
     
     
     if paired:
       ## Add for paired
       sXY = np.nansum((X - mX)*(Y - mY), axis=1)
       
       ## Number of not NA values in each row
       n = np.sum(~np.isnan(X*Y), axis=1)
       
       ## d == difference between the group means for each row (==gene)
       ## s == pooled standard deviation for each row (==gene)        
       d = mY - mX
       s = np.sqrt(((sX + sY) / (n + n - 2)) * (2 / n) - 2/(n*n-n)*sXY)
       
       ## Cases with less than two non-missing values.
       ## Set d = 0, s = 1
       ind = np.where( n < 2 )
       d[ind] = 0
       s[ind] = 1
     
     return  {'d': d.reshape(-1), 's': s.reshape(-1)}
   
   # Multiple groups
   if len(samples)>2:
     
     samples_all = np.concatenate(samples, axis=1) # do.call("cbind",samples)
     
     if not paired:
       sum_cols = np.sum([sample.shape[1] for sample in samples])
       prod_cols = np.prod([sample.shape[1] for sample in samples])
       f = sum_cols / prod_cols #f <- sum(sapply(samples, ncol)) / prod(sapply(samples, ncol))

       r = np.zeros(samples_all.shape[0]) #r <- vector(mode="numeric", length=nrow(samples.all))
       for k in range(len(samples)):
         r = r + (np.nanmean(samples[k], axis=1)-np.nanmean(samples_all, axis=1))**2 #r <- r + (rowMeans(samples[[k]], na.rm=TRUE)-rowMeans(samples.all, na.rm=TRUE))^2       
       d = (f*r)**0.5
       
       f = 1/np.sum([sample.shape[1] for sample in samples]-1) * np.sum(1/[sample.shape[1] for sample in samples]) #f <- 1/sum(sapply(samples, ncol)-1) * sum(1/sapply(samples, ncol))
       s = np.zeros(samples_all.shape[0])  # s <- vector(mode="numeric", length=nrow(samples.all))
       for k in range(len(samples)):
         s = s + np.sum(np.apply_along_axis(lambda x: (x - np.nanmean(x))**2, axis=1, arr=samples[k,:]), axis=0, where=~np.isnan(samples[k,:]).any(axis=0)) #s <- s + colSums(apply(samples[[k]], 1, function(x) (x-mean(x,na.rm=TRUE))^2), na.rm=TRUE)       
       s = (f*s)**0.5       
     
     if paired:
       raise ValueError("Multiple paired groups not supported!")     
     
     return {'d': d.reshape(-1), 's': s.reshape(-1)}

def calculateP(observed, permuted):  
  # Store order for later use
  observed_order = sorted(range(len(observed)), key=lambda k: abs(observed[k]), reverse=True) # order(abs(observed), decreasing=TRUE)
  
  # Sort observed and permuted values
  observed = -np.sort(-abs(observed)) #sort(abs(observed), decreasing=TRUE)
  permuted = -np.sort(-np.abs(permuted.flatten())) #sort(abs(as.vector(permuted)), decreasing=TRUE)
  
  # Get p-values from C++ code
  # (expects ordered vectors)
  p = pvalue(observed, permuted)        
  
  # Revert to original ordering
  results = np.zeros(len(p)) #vector(mode="numeric", length=length(p))
  for i in range(len(results)):
    results[observed_order[i]] =  p[i]
  
  return results


def calculateFDR(observed, permuted, progress):
  observed = abs(observed)
  permuted = abs(permuted)
  ord = np.argsort(-observed, kind='mergesort') #order(observed, decreasing=TRUE, na.last=TRUE)
  a = observed[ord]
  
  A = np.empty((len(a), permuted.shape[1])) #matrix(NA, nrow=length(a), ncol=ncol(permuted))
  A[:] = np.nan
  if progress:
    pb = tqdm(total=A.shape[1]) #txtProgressBar(min=0, max=ncol(A), style=3)
  for i in range(A.shape[1]): #seq_len(ncol(A))
    sorted_column = np.sort(permuted[:,i])[::-1] #sort(permuted[,i], decreasing=TRUE, na.last=TRUE)
    a_rand = np.concatenate([sorted_column[~np.isnan(sorted_column)], sorted_column[np.isnan(sorted_column)]])
    n_bigger = biggerN(a, a_rand)
    A[ord,i] = n_bigger/range(1, len(a)+1)
    if progress:
      pb.update()      

  if progress:
    pb.close()
  
  
  FDR = np.apply_along_axis(np.median, 1, A) #apply(A, 1, median)
  FDR[FDR>1] = 1
  FDR[ord] = list(reversed([min(FDR[ord][x-1:]) for x in range(len(FDR), 0, -1)])) #rev(sapply(length(FDR):1, function(x) return(min(FDR[ord][x:length(FDR)]))))

  return FDR


def biggerN(x, y):  
  sorted_x = np.sort(x)[::-1]
  x = np.concatenate([sorted_x[~np.isnan(sorted_x)], sorted_x[np.isnan(sorted_x)]]) #sort(x, decreasing=TRUE, na.last=TRUE)		# sort x in decreasing order
  sorted_y = np.sort(y)[::-1]
  y = np.concatenate([sorted_y[~np.isnan(sorted_y)], sorted_y[np.isnan(sorted_y)]]) #sort(y, decreasing=TRUE, na.last=TRUE)		# sort y in decreasing order
  #a <- match(x, x)				 # vector of the positions of (first) matches of the first argument in the second
  a = np.array([np.where(x == v)[0][0] if v in x else None for v in x])
  #b <- x %in% y				   # a logical vector indicating if there is a match or not for its left operand
  b = np.isin(x, y)
  sorted_z = np.sort(np.concatenate([x, y]))[::-1]
  z = np.concatenate([sorted_z[~np.isnan(sorted_z)], sorted_z[np.isnan(sorted_z)]]) #sort(c(x, y), decreasing=TRUE, na.last=TRUE)		# sort c(x,y) in decreasing order
  #d <- match(x, z)				 # vector of the positions of (first) matches of the first argument in the second
  # z_indices = np.argsort(z)
  # match_indices = np.searchsorted(z[z_indices], x, side='left')
  # matches = np.full(x.shape, np.nan)
  # matches[np.argsort(z_indices)] = z_indices[match_indices]
  # d = matches
  #ipdb.set_trace(context=10)   ## BREAKPOINT
  d = np.array([np.where(z == v)[0][0] if v in z else None for v in x])
  res = d - a + b

  return res

@njit
def pvalue(a, b):
  observed = a.ravel()
  permuted = b.ravel()
  pvalues = np.zeros(len(observed))

  j = 0
  for i in range(len(observed)):
    while permuted[j] >= observed[i] and j < len(permuted):
      j += 1
    pvalues[i] = float(j) / len(permuted)

  return pvalues.reshape(a.shape)


import numpy as np
from numba import njit, jit
import pandas as pd


def calculateOverlaps1(D, S, pD, pS, D_len, N, N_len, ssq_i, B, overlaps, overlaps_P):
  D_ovlp = D.ravel()
  S_ovlp = S.ravel()
  pD_ovlp = pD.ravel()
  pS_ovlp = pS.ravel()
  overlaps_ovlp = overlaps.ravel()
  overlaps_P_ovlp = overlaps_P.ravel()
    
  res1 = np.zeros((D_len,1))
  res2 = np.zeros((D_len,1))
  pres1 = np.zeros((D_len,1))
  pres2 = np.zeros((D_len,1))
  for b in range(1, B):
    for i in range(D_len):
      
      res1[i] = np.abs( D_ovlp[(b-1) * D_len + i] / ( S_ovlp[(b-1) * D_len + i] + ssq_i) )
      res2[i] = np.abs( D_ovlp[(b + B - 1) * D_len + i] / ( S_ovlp[(b + B - 1) * D_len + i] + ssq_i) )
      pres1[i] = np.abs( pD_ovlp[(b-1) * D_len + i] / (pS_ovlp[(b-1) * D_len + i] + ssq_i))
      pres2[i] = np.abs( pD_ovlp[(b + B - 1) * D_len + i] / (pS_ovlp[(b + B -1) * D_len + i] + ssq_i))    

    overlaps_ovlp = calculateOverlap_1(res1, res2, D_len, N, N_len, b, B, overlaps_ovlp)
    ##### CalculateOverlap_1: overlaps ####  overlaps_res = calculateOverlap_1(res1, res2, D_len, N, N_len, b, B, overlaps_ovlp)    
    # Copy r2 and sort the copy.
    # r3 = sorted(res2, reverse=True)
    # # Sort r2 by r1
    # r1, r2 = sort2_1(res1, res2, D_len)
    # #Calculate the overlap
    # for k in range(N_len):
    #   sum = 0
    #   for j in range(N[k]):
    #     sum += (r2[j] <= r3[N[k] - 1])
    #   overlaps_ovlp[ (b-1) + k*B ] = sum / N[k]
    
    #del r3
    #########################
    
    overlaps_P_ovlp = calculateOverlap_1(pres1, pres2, D_len, N, N_len, b, B, overlaps_P_ovlp)
    ##### CalculateOverlap_1: overlaps_P ####  overlaps_P_res = calculateOverlap_1(pres1, pres2, D_len, N, N_len, b, B, overlaps_P_ovlp)
    # Copy r2 and sort the copy.
    # pr3 = sorted(pres2, reverse=True)    
    # # Sort r2 by r1
    # pr1, pr2 = sort2_1(pres1, pres2, D_len)    
    # #Calculate the overlap
    # for k in range(N_len):
    #   sum = 0
    #   for j in range(N[k]):
    #     sum += (pr2[j] <= pr3[N[k] - 1])
    #   overlaps_P_ovlp[ (b-1) + k*B ] = sum / N[k]
    
    #del pr3
    #########################  

  return {'overlaps': overlaps_ovlp.reshape(overlaps.shape), 'overlaps_P': overlaps_P_ovlp.reshape(overlaps_P.shape)} 

# Calculate the overlap

def calculateOverlap_1(r1, r2, r_len, N, N_len, b, B, overlaps):
  # Copy r2 and sort the copy.
  r3 = sorted(r2, reverse=True)
  
  # Sort r2 by r1
  r1, r2 = sort2_1(r1, r2, r_len)
  
  #Calculate the overlap
  for i in range(N_len):
    sum = 0
    for j in range(N[i]):
      sum += (r2[j] <= r3[N[i] - 1])
    overlaps[ (b-1) + i*B ] = sum / N[i]

  return overlaps

# Sort array b based on the array a (decreasingly)

def sort2_1(a, b, n):
  pairs = [] #np.zeros([n, 2], dtype=float, order='C')
  for i in range(n):
    pairs.append((a[i], b[i]))
  
  # Sort the pairs (inc)
  pairs = sorted(pairs)

  # Split the pairs back into the original vectors (dec).
  for i in range(n):
    a[n-1-i] = pairs[i][0]
    b[n-1-i] = pairs[i][1]
  
  #del pairs

  return (a, b)